mutation($executionParams: ExecutionParams!) {
  launchPipelineExecution(executionParams: $executionParams) {
    __typename
    ... on LaunchPipelineRunSuccess {
      run {
        runId
        pipeline {
          ... on PipelineReference {
            name
          }
        }
      }
    }
    ... on PipelineConfigValidationInvalid {
      pipelineName
      errors {
        message
      }
    }
    ... on PipelineNotFoundError {
      pipelineName
    }
  }
}
